from functools import cached_property
from qiskit.quantum_info import Statevector
# from qiskit.opflow import CircuitStateFn # old qiskit function replaced by Statevector
from qiskit import QuantumCircuit
from symmer import process, QuantumState, PauliwordOp
from symmer.operators.utils import (
    symplectic_to_string, safe_PauliwordOp_to_dict, safe_QuantumState_to_dict
)
from symmer.evolution import PauliwordOp_to_QuantumCircuit, get_CNOT_connectivity_graph, topology_match_score
from networkx.algorithms.cycles import cycle_basis
from scipy.optimize import minimize
from scipy.sparse import csc_array
from copy import deepcopy
import numpy as np
from typing import *

class VQE_Driver:
    """
    expectation value method one of the following choices:
        - symbolic_direct: uses symmer to compute <psi|H|psi> directly using the QuantumState class
        - symbolic_projector: computes expval by projecting onto +-1 eigenspaces of observable terms
        - observable_rotation: implements the circuit as rotations applied to the observable
        - sparse_array: direct calcaultion by converting observable/state to sparse array
        - dense_array: direct calcaultion by converting observable/state to dense array
    
    Attributes:
        expectation_eval (str): expectation value method. Its default value is 'symbolic_direct'.
        verbose (bool): If True, prints out useful information during computation. By default it is set to 'True'.
    """
    expectation_eval = 'sparse_array'
    # prints out useful information during computation:
    verbose = True
       
    def __init__(self,
        observable: PauliwordOp,
        ansatz_circuit: QuantumCircuit = None,
        excitation_ops: PauliwordOp = None,
        ref_state: QuantumState = None
        ) -> None:
        """
        Args:
            observable (PauliwordOp): Observables
            ansatz_circuit (QuantumCircuit): Ansatz Circuit. By default, it's set to 'None'.
            excitation_ops (PauliwordOp): Excitation Operators. By default, it's set to 'None'.
            ref_state (QuantumState): Reference State. By default, it's set to 'None'.
        """
        self.observable = observable
        self.ref_state = ref_state
        # observables must have real coefficients over the Pauli group:
        assert np.all(self.observable.coeff_vec.imag == 0), 'Observable not Hermitian'
        
        if excitation_ops is not None:
            self.prepare_for_evolution(excitation_ops)
        else:
            self.circuit = ansatz_circuit

    def prepare_for_evolution(self, excitation_ops: PauliwordOp) -> None:
        """ 
        Save the excitation generators and construct corresponding ansatz circuit.

        Args:
            excitation_ops (PauliwordOp): Excitation Operators.
        """
        self.excitation_generators = PauliwordOp(
            excitation_ops.symp_matrix, np.ones(excitation_ops.n_terms)
        )
        self.circuit = PauliwordOp_to_QuantumCircuit(
                PwordOp=self.excitation_generators, ref_state=self.ref_state, bind_params=False
        )

    def get_state(self, 
            evolution_obj: Union[QuantumCircuit, PauliwordOp], 
            x: np.array
        ) -> Union[np.array, QuantumState, List[Tuple[PauliwordOp, float]]]:
        """ 
        Args:
            evolution_obj (Union[QuantumCircuit, PauliwordOp]): Evoluation object is either a Quantum Circuit (QuantumCircuit) or a Excitation Generating Set(PauliwordOp).

        Returns:
            Given a quantum circuit or excitation generating set, return the relevant state-type object:
            - Array (np.array) of the QuantumCircuit (for sparse/dense array methods).
            - Quantum State (QuantumState) representation of the QuantumCircuit (for symbolic methods).
            - Rotations of the form [(generator, angle)] for the observable_rotation expectation_eval method.
        """
        if self.expectation_eval == 'observable_rotation':
            return list(zip(evolution_obj, -2*x))
        else:
            state = Statevector(evolution_obj.assign_parameters(x)).data.reshape([-1,1])
            if self.expectation_eval == 'dense_array':
                return state
            elif self.expectation_eval == 'sparse_array':
                return csc_array(state)
            elif self.expectation_eval.find('symbolic') != -1:
                return QuantumState.from_array(state)
        
    def _f(self, 
           observable: PauliwordOp, 
           state: Union[np.array, QuantumState, List[Tuple[PauliwordOp, float]]]
        ) -> float:
        """ 
        Given an observable and state in the relevant form for the
        expectation value method, calculate the expectation value and return.

        Args:
            observable (PauliwordOp): Observable
            state: State-type object. It can be:
                - Array (np.array) of the QuantumCircuit (for sparse/dense array methods).
                - Quantum State (QuantumState) representation of the QuantumCircuit (for symbolic methods).
                - Rotations of the form [(generator, angle)] for the observable_rotation expectation_eval method.
        
        Returns:
            Expectation Value (float)
        """
        if self.expectation_eval == 'dense_array':
            return (state.conjugate().T @ observable.to_sparse_matrix.toarray() @ state)[0,0].real
        elif self.expectation_eval == 'sparse_array':
            return (state.conjugate().T @ observable.to_sparse_matrix @ state)[0,0].real
        elif self.expectation_eval == 'symbolic_projector':
            return observable.expval(state).real
        elif self.expectation_eval == 'symbolic_direct':
            return (state.dagger * observable * state).real   
        elif self.expectation_eval == 'observable_rotation':
            return (self.ref_state.dagger * observable.perform_rotations(state) * self.ref_state).real
        
    def f(self, x: np.array) -> float:
        """ 
        Given a parameter vector, bind to the circuit and retrieve expectation value.

        Args:
            x (np.array): Parameter vector

        Returns:
            Expectation Value (float)
        """
        if self.expectation_eval == 'observable_rotation':
            state = self.get_state(self.excitation_generators, x)
        else:
            state = self.get_state(self.circuit, x)
        return self._f(self.observable, state)
        
    def partial_derivative(self, x: np.array, param_index: int) -> float:
        """ 
        Get the partial derivative with respect to an ansatz parameter
        by the parameter shift rule.

        Args:
            x (np.array): Parameter vector
            param_index (int): Prarameter index
        
        Returns:
            Partial derivative(float) with respect to an ansatz parameter.
        """
        x_upper = x.copy(); x_upper[param_index]+=np.pi/4
        x_lower = x.copy(); x_lower[param_index]-=np.pi/4
        return self.f(x_upper) - self.f(x_lower)
    
    def gradient(self, x: np.array) -> np.array:
        """ 
        Get the ansatz parameter gradient, i.e. the vector of partial derivatives.

        Args:
            x (np.array): Parameter vector
        
        Returns:
            Ansatz parameter gradient (np.array)
        """
        if self.expectation_eval.find('projector') == -1:
            @process.parallelize
            def f(index, param):
                return self.partial_derivative(param,index)
            grad_vec = f(range(self.circuit.num_parameters), x)

        else:
            grad_vec = [self.partial_derivative(x, i) for i in range(self.circuit.num_parameters)]
        
        return np.asarray(grad_vec)
    
    def run(self, x0:np.array=None, **kwargs):
        """ 
        Run the VQE routine.
        
        Args:
            x0 (np.array): Parameter vector
        """
        if x0 is None:
            x0 = np.random.random(self.circuit.num_parameters)
        
        vqe_history = {'params':{}, 'energy':{}, 'gradient':{}}

        # set up a counter to keep track of optimization steps this is important 
        # as some optimizers do not compute gradients at each optimization step and 
        # therefore must be labeled to match with the correct iteration later on
        global counter
        counter = -1
        def get_counter(increment=True):
            global counter
            if increment:
                counter += 1
            return counter

        # wrap VQE_Driver.f() for the optimizer and store the interim values
        def fun(x):    
            counter = get_counter(increment=True)
            energy  = self.f(x)
            vqe_history['params'][counter] = tuple(x)
            vqe_history['energy'][counter] = energy
            if self.verbose:
                print(f'Optimization step {counter: <2}:\n\t Energy = {energy}')
            return energy

        # wrap VQE_Driver.gradient() for the optimizer and store the interim values
        def jac(x):
            counter = get_counter(increment=False)
            grad    = self.gradient(x)
            vqe_history['gradient'][counter] = tuple(grad)
            if self.verbose:
                print(f'\t    |∆| = {np.linalg.norm(grad)}')
            return grad
        
        if self.verbose:
            print('VQE simulation commencing...\n')
        opt_out = minimize(
            fun=fun, jac=jac, x0=x0, **kwargs
        )
        return serialize_opt_data(opt_out), vqe_history

class ADAPT_VQE(VQE_Driver):
    """ 
    Performs qubit-ADAPT-VQE (https://doi.org/10.1103/PRXQuantum.2.020310), a 
    variant of ADAPT-VQE (https://doi.org/10.1038/s41467-019-10988-2) that takes 
    its excitation pool as Pauli operators (mapped via some transformation such 
    as Jordan-Wigner) instead of the originating fermionic operators.

    Attributes:
        derivative_eval (str): Method which is to be used to calculate the operator pool derivatives.
        TETRIS (bool): If True, TETRIS-ADAPT-VQE is performed. By default it is set to False.
        topology_aware (bool): If True, Hardware-Aware ADAPT-VQE is performed. By default it is set to True.
        topology_bias (float): Bias value used in Hardware-Aware ADAPT-VQE. It's default value is 1.
    """
    # method by which to calculate the operator pool derivatives, either
    # commutators: compute the commutator of the observable with each pool element
    # param_shift: use the parameter shift rule, requiring two expectation values per derivative
    derivative_eval = 'param_shift'
    # we have alost implemented TETRIS-ADAPT-VQE as per https://doi.org/10.48550/arXiv.2209.10562
    # that aims to reduce circuit-depth in the ADAPT routine by adding multiple excitation terms
    # per cycle that are supported on distinct qubit positions.
    TETRIS = False
    topology_aware = False
    topology_bias = 1
    topology = None
    subgraph_match_depth = 3
    
    def __init__(self,
        observable: PauliwordOp,
        excitation_pool: PauliwordOp = None,
        ref_state: QuantumState = None
        ) -> None:
        """
        Args:
            observable (PauliwordOp): Observable
            excitation_pool (PauliwordOp): Excitation pool as Pauli operators. By default, it's set to 'None'.
            ref_state (QuantumState): Reference State. By default, it's set to 'None'.
        """
        super().__init__(
            observable     = observable,
            excitation_ops = PauliwordOp.empty(observable.n_qubits),
            ref_state      = ref_state
        )
        self.excitation_pool = PauliwordOp(
            excitation_pool.symp_matrix, np.ones(excitation_pool.n_terms)
        )
        self.adapt_operator = PauliwordOp.empty(observable.n_qubits)
        self.opt_parameters = []
        self.current_state  = None
      
    @cached_property
    def commutators(self) -> List[PauliwordOp]:
        """ 
        List of commutators [H, P] where P is some operator pool element.

        Returns:
            List of commutators [H, P]
        """
        @process.parallelize
        def f(P, obs):
            return obs.commutator(P)*1j
        commutators = f(self.excitation_pool, self.observable)
        return commutators
        
    def _derivative_from_commutators(self, index: int) -> float:
        """ 
        Calculate derivative using the commutator method.

        Args:
            index (int): Index

        Returns:
            Derivative (float) using the commutator method.
        """
        assert self.current_state is not None
        return self._f(observable=self.commutators[index], state=self.current_state) 
    
    def _derivative_from_param_shift(self, index):
        """ 
        Calculate the derivative using the parameter shift rule.

         Args:
            index (int): Index

        Returns:
            Derivative (float) using the parameter shift rule.
        """
        adapt_op_temp = self.adapt_operator.append(self.excitation_pool[index])
        circuit_temp = PauliwordOp_to_QuantumCircuit(
            PwordOp=adapt_op_temp, ref_state=self.ref_state, bind_params=False)
        upper_state = self.get_state(circuit_temp, np.append(self.opt_parameters, +np.pi/4))
        lower_state = self.get_state(circuit_temp, np.append(self.opt_parameters, -np.pi/4))
        return self._f(self.observable,upper_state) - self._f(self.observable,lower_state)

    def pool_gradient(self):
        """ 
        Get the operator pool gradient by calculating the derivative with respect to
        each element of the pool. This is parallelized for all but the symbolic_projector
        expectation value calculation method as that is already multiprocessed and therefore
        would result in nested daemonic processes.

        Returns:
            Operator pool gradient (np.array)
        """
        if self.derivative_eval == 'commutators':
            self.commutators # to ensure this has been cached, else nested daemonic process occurs            
            if self.expectation_eval == 'observable_rotation':
                self.current_state = self.get_state(self.adapt_operator, self.opt_parameters)
            else:
                circuit_temp = PauliwordOp_to_QuantumCircuit(
                    PwordOp=self.adapt_operator, ref_state=self.ref_state, bind_params=False)
                self.current_state = self.get_state(circuit_temp, self.opt_parameters)
            if self.expectation_eval in ['sparse_array', 'symbolic_direct', 'observable_rotation']:
                # the commutator method may be parallelized since the state is constant
                @process.parallelize
                def f(index, obs):
                    return obs._derivative_from_commutators(index)
                gradient = f(range(self.excitation_pool.n_terms), self)
            else:
                # ... unless using symbolic_projector since this is multiprocessed
                gradient = list(map(self._derivative_from_commutators, range(self.excitation_pool.n_terms)))
        
        elif self.derivative_eval == 'param_shift':
            # not parallelizable due the Statevector already using multiprocessing! 
            gradient = list(map(self._derivative_from_param_shift, range(self.excitation_pool.n_terms)))
        
        else:
            raise ValueError('Unrecognised derivative_eval method')
        
        return np.asarray(gradient)
    
    def pool_score(self):
        """ 
        Score the operator pool with respect to gradients and topology likeness.
        """
        scores = abs(self.pool_gradient())

        if self.topology_aware:
            assert self.topology is not None, 'No hardware topology specified'
            # Hardware-Aware ADAPT-VQE favours circuits that match the target topology closely
            topology_scores = []
            for index in range(self.excitation_pool.n_terms):
                adapt_op_temp = self.adapt_operator.append(self.excitation_pool[index])
                topology_scores.append(
                    topology_match_score(adapt_op_temp, self.topology, max_depth=self.subgraph_match_depth)
                )
            scores *= np.power(np.array(topology_scores), self.topology_bias)
        
        return scores
        
    def append_to_adapt_operator(self, excitations_to_append: List[PauliwordOp]):
        """ 
        Append the input term(s) to the expanding adapt_operator.
        """
        for excitation in excitations_to_append:
            if ~np.any(self.adapt_operator.symp_matrix):
                self.adapt_operator += excitation
            else:
                self.adapt_operator = self.adapt_operator.append(excitation)
        
    def optimize(self, 
            max_cycles:int=10, gtol:float=1e-3, atol:float=1e-10, 
            target:float=0, target_error:float=1e-3
        ):
        """ 
        Perform the ADAPT-VQE optimization

        Args:
            gtol: gradient throeshold below which optimization will terminate
            atol: if the difference between successive expectation values is below this threshold, terminate
            max_cycles: maximum number of ADAPT cycles to perform
            target: if a target energy is known, this may be specified here
            target_error: the absoluate error threshold with respect to the target energy 
        """
        interim_data = {'history':[]}
        adapt_cycle=1
        gmax=1
        anew=1
        aold=0
        
        while (
                gmax>gtol and adapt_cycle<=max_cycles and 
                abs(anew-aold)>atol and abs(anew-target)>target_error
            ):
            # save the previous gmax to compare for the gdiff check
            aold = deepcopy(anew)
            # calculate gradient across the pool and select term with the largest derivative
            scores = self.pool_score()
            grad_rank = list(map(int, np.argsort(scores)[::-1]))
            gmax = scores[grad_rank[0]]

            # TETRIS-ADAPT-VQE
            if self.TETRIS:
                new_excitation_list = []
                support_mask = np.zeros(self.observable.n_qubits, dtype=bool)
                for i in grad_rank:
                    new_excitation = self.excitation_pool[i]
                    support_exists = (new_excitation.X_block | new_excitation.Z_block) & support_mask
                    if ~np.any(support_exists):
                        new_excitation_list.append(new_excitation)
                        support_mask = support_mask | (new_excitation.X_block | new_excitation.Z_block)
                    if np.all(support_mask) or scores[i] < gtol:
                        break
            else:
                new_excitation_list = [self.excitation_pool[grad_rank[0]]]
                
            # append new term(s) to the adapt_operator that stores our ansatz as it expands
            n_new_terms = len(new_excitation_list)
            self.append_to_adapt_operator(new_excitation_list)
                        
            if self.verbose:
                print('-'*39)
                print(f'ADAPT cycle {adapt_cycle}\n')
                print(f'Largest pool derivative ∂P∂θ = {gmax: .5f}\n')
                print('Selected excitation generator(s):\n')
                for op in new_excitation_list:
                    print(f'\t{symplectic_to_string(op.symp_matrix[0])}')
                print('\n', '-'*39)
            
            # having selected a new term to append to the ansatz, reoptimize with VQE
            self.prepare_for_evolution(self.adapt_operator)
            opt_out, vqe_hist = self.run(
                x0=np.append(self.opt_parameters, [0]*n_new_terms), method='BFGS'
            )
            interim_data[adapt_cycle] = {
                'output':opt_out, 'history':vqe_hist, 'gmax':gmax, 
                'excitation': [symplectic_to_string(t.symp_matrix[0]) for t in new_excitation_list]
            }
            anew = opt_out['fun']
            interim_data['history'].append(anew)
            if self.verbose:
                print(F'\nEnergy at ADAPT cycle {adapt_cycle}: {anew: .5f}\n')
            self.opt_parameters = opt_out['x']
            adapt_cycle+=1

        return {
            'result': opt_out, 
            'interim_data': interim_data,
            'ref_state': safe_QuantumState_to_dict(self.ref_state),
            'adapt_operator': [symplectic_to_string(t) for t in self.adapt_operator.symp_matrix]
        }
    
def serialize_opt_data(opt_data):
    return {
        'message':opt_data.message, 'success':opt_data.success, 'status':opt_data.status,
        'fun':opt_data.fun, 'x':tuple(opt_data.x),'jac':tuple(opt_data.jac),
        'nit':opt_data.nit, 'nfev':opt_data.nfev,'njev':opt_data.njev,
    }


from functools import reduce
from typing import Dict, List, Union
from symmer.operators import PauliwordOp, QuantumState
from symmer.evolution.gate_library import *
from qiskit.circuit import QuantumCircuit, ParameterVector
from networkx import Graph, draw_spring
from collections import Counter
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning) 

##############################################
# Decompose any QASM file into a PauliwordOp #
##############################################

def qasm_to_PauliwordOp(qasm: str, reverse=False, combine=True) -> PauliwordOp:
    """ 
    Decompose an QASM circuit into a linear combination of Pauli 
    operators via the gate definitions in evolution.gate_library.

    Args:
        qasm (str): QASM circuit which you want to convert into  a linear combination of Pauli operators.
        reverse (bool): If 'True' then flip qubit ordering and negate angles (for consistency with Qiskit). By default it's set to 'False'.
        combine (bool): If 'True' then take product over gateset - obscures gate contributions in resulting PauliwordOp. By default it's set to 'True'.

    Returns:
        A linear combination of Pauli operators representing QASM circuit.
    """
    gate_map = {
        'x':X, 'y':Y, 'z':Z, 'h':Had, 'rx':RX, 'ry':RY, 
        'rz':RZ, 'u1':U1, 'cz':CZ, 'cx':CX, 's':S
    } # for conversion from qiskit to PauliwordOp definitions
    gateset = []
    for gate in qasm.split(';\n')[:-1]:
        name, qubits = gate.split(' ')
        # identify number of qubits in circuit
        if name=='qreg':
            num_qubits = int(qubits[2:-1])
        if name in ['barrier', 'include', 'OPENQASM', 'qreg']:
            pass
        else:
            # extract angle
            if name.find('(')!=-1:
                name, angle = name.split('(')
                angle = angle[:-1]
                if angle=='pi/2':
                    angle = np.pi/2
                elif angle=='-pi/2':
                    angle = -np.pi/2
                else:
                    angle = float(angle)
            else:
                angle = None
            #extract qubits
            if qubits.find(',')!=-1:
                control, target = qubits.split(',')
                control, target = int(control[2:-1]), int(target[2:-1])
            else:
                control, target = -1, int(qubits[2:-1])
            # if reverse then flip qubit ordering and negate angles (for consistency with Qiskit)
            flip=1
            if reverse:
                flip=-1
                control, target = num_qubits-1-control, num_qubits-1-target
            # generate relevant gate and append to list
            if name in ['x', 'y', 'z', 'h', 's', 'sdg']:
                G = gate_map[name](num_qubits, target)
            elif name in ['cz', 'cx']:
                G = gate_map[name](num_qubits, control, target)
            elif name in ['rx', 'ry', 'rz', 'u1']:
                G = gate_map[name](num_qubits, target, angle=flip*angle)
            else:
                raise ValueError(f'Gate decomposition {name} not defined')
            gateset.append(G)

    # if combine then take product over gateset - obscures gate contributions in resulting PauliwordOp
    if combine:
        qc_decomposition = reduce(lambda x,y:x*y, gateset[::-1])
        return qc_decomposition.cleanup()
    else:
        return gateset

####################################################
# Trotterized circuit of exponentiated PauliwordOp #
####################################################

def PauliwordOp_to_instructions(PwordOp) -> Dict[int, Dict[str, List[int]]]:
        """ 
        Stores a dictionary of gate instructions at each step, where each value
        is a dictionary indicating the indices on which to apply each H,S,CNOT and RZ gate.

        Args:
            PwordOp (PauliWordOp): PauliwordOp which has to be converted into Trotterized circuit instructions.

        Returns:
            circuit_instructions (dict): Trotterized circuit
        """
        circuit_instructions = {}
        for step, (X,Z) in enumerate(zip(PwordOp.X_block, PwordOp.Z_block)):
            # locations for H and S gates to transform into Pauli Z basis
            H_indices = np.where(X)[0][::-1]
            S_indices = np.where(X & Z)[0][::-1]
            # CNOT cascade indices
            CNOT_indices = np.where(X | Z)[0][::-1]
            circuit_instructions[step] = {'H_indices':H_indices, 
                                        'S_indices':S_indices, 
                                        'CNOT_indices':CNOT_indices,
                                        'RZ_index':CNOT_indices[-1]}
        return circuit_instructions

def PauliwordOp_to_QuantumCircuit(
    PwordOp: PauliwordOp, 
    ref_state: np.array  = None,
    basis_change_indices: Dict[str, List[int]] = {'X_indices':[],'Y_indices':[]},
    trotter_number: int = 1, 
    bind_params: bool = True,
    include_barriers:bool = True,
    parameter_label: str = 'P'
    ) -> QuantumCircuit:
    """
    Convert the operator to a QASM circuit string for input 
    into quantum computing packages such as Qiskit and Cirq.

    Args:
        PwordOp (PauliwordOp): PauliwordOp which has to be converted into Quantum Circuit.
        ref_state (np.array): Reference State. By default, it is set to None.
        basis_change_indices (Dict[str, List[int]]): Basis change indices in form [X_indices, Y_indices].
        trotter_number (int): Trotter Number. By default, it's set to 1.
        bind_params (bool): If 'True', parameters are binded. By default it's set to 'True'.
        include_barriers (bool): If 'True', the quantum circuit includes barriers. By default it's set to 'True'.
        parameter_label (str): Parameter label. By default it's set to 'P'.

    Returns:
        Quantum Circuit representation of Operator
    """
    if isinstance(ref_state, QuantumState):
        assert ref_state.n_terms == 1
        ref_state = ref_state.state_matrix[0]

    def qiskit_ordering(indices):
        """ 
        We index from left to right - in Qiskit this ordering is reversed.

        Args:
            indices (int): Indices
        """
        return PwordOp.n_qubits - 1 - indices

    qc = QuantumCircuit(PwordOp.n_qubits)
    for i in qiskit_ordering(np.where(ref_state==1)[0]):
        qc.x(i)

    non_identity = PwordOp[np.any(PwordOp.symp_matrix, axis=1)]

    if non_identity.n_terms > 0:

        def CNOT_cascade(cascade_indices, reverse=False):
            index_pairs = list(zip(cascade_indices[:-1], cascade_indices[1:]))
            if reverse:
                index_pairs = index_pairs[::-1]
            for source, target in index_pairs:
                qc.cx(source, target)

        def circuit_from_step(angle, H_indices, S_indices, CNOT_indices, RZ_index):
            # to Pauli X basis
            for i in S_indices:
                qc.sdg(i)
            # to Pauli Z basis
            for i in H_indices:
                qc.h(i)
            # compute parity
            CNOT_cascade(CNOT_indices)
            qc.rz(-2*angle, RZ_index)
            CNOT_cascade(CNOT_indices, reverse=True)
            for i in H_indices:
                qc.h(i)
            for i in S_indices:
                qc.s(i)

        if bind_params:
            angles = non_identity.coeff_vec.real/trotter_number
        else:
            angles = np.array(ParameterVector(parameter_label, non_identity.n_terms))/trotter_number

        instructions = PauliwordOp_to_instructions(non_identity)
        assert(len(angles)==len(instructions)), 'Number of parameters does not match the circuit instructions'
        for trot_step in range(trotter_number):
            for step, gate_indices in instructions.items():
                qiskit_gate_indices = [qiskit_ordering(indices) for indices in gate_indices.values()]

                if include_barriers:
                    qc.barrier()

                circuit_from_step(angles[step], *qiskit_gate_indices)

    if include_barriers:
        qc.barrier()

    for i in basis_change_indices['Y_indices']:
        qc.s(qiskit_ordering(i))
    for i in basis_change_indices['X_indices']:
        qc.h(qiskit_ordering(i))
        
    return qc


import warnings

import numpy as np
from typing import List, Union
from functools import cached_property
from symmer.projection import S3Projection
from symmer.operators import PauliwordOp, IndependentOp, QuantumState

class QubitTapering(S3Projection):
    """ 
    Class for performing qubit tapering as per https://arxiv.org/abs/1701.08213.
    Reduces the number of qubits in the problem whilst preserving its energy spectrum by:

    1. identifying a symmetry of the Hamiltonian,
    2. finding an independent basis therein,
    3. rotating each basis operator onto a single Pauli X, 
    4. dropping the corresponding qubits from the Hamiltonian whilst
    5. fixing the +/-1 eigenvalues

    Steps 1-2 are handled in this class whereas we defer to the parent S3Projection for 3-5.
    """
    name = 'qubit_tapering' # for reference in QubitSubspaceManager

    def __init__(self,
            operator: PauliwordOp, 
            target_sqp: str = 'Z'
        ) -> None:
        """ 
        Input the PauliwordOp we wish to taper.
        There is freedom over the choice of single-qubit Pauli operator we wish to rotate onto, 
        however this is set to X by default (in line with the original tapering paper).

        Args: 
            operator (PauliwordOp): The Operator you want to tapper.
            target_sqp (str): The single-qubit Pauli operator we wish to rotate onto. By default, it is set to 'X'.
        """
        self.operator = operator
        self.target_sqp = target_sqp
        self.n_taper = self.symmetry_generators.n_terms
        super().__init__(self.symmetry_generators)
        
    @cached_property
    def symmetry_generators(self) -> IndependentOp:
        """ 
        Find an independent basis for the input operator symmetry.

        Returns:
            stabilizers (IndependentOp): Set of stablizers.
        """
        stabilizers = IndependentOp.symmetry_generators(self.operator)
        stabilizers.target_sqp = self.target_sqp
        return stabilizers

    def taper_it(self,
            ref_state: Union[List[int], np.array]=None,
            sector: Union[List[int], np.array]=None,
            aux_operator: PauliwordOp = None
        ) -> PauliwordOp:
        """ 
        Finally, once the symmetry generators and sector have been identified, 
        we may perform a projection onto the corresponding stabilizer subspace via 
        the parent S3Projection class.

        This method allows one to input an auxiliary operator other than the internal
        operator itself to be tapered consistently with the identified symmetry. This is 
        especially useful when considering an Ansatz defined over the full system that 
        one wishes to restrict to the same stabilizer subspace as the Hamiltonian for 
        use in VQE, for example.

        Args:
            ref_state (np.array): Reference State. By default, it is set to None.
            sector (np.array): Sector. By default it is set to none.
            aux_operator (PauliwordOp): Auxiliary operator. By default, it is set to None.

        Returns:
            tapered_operator (PauliwordOp): Tappered Operator
        """
        if ref_state is not None:
            if not isinstance(ref_state, QuantumState):
                ref_state = QuantumState(ref_state)
            assert ref_state._is_normalized(), 'Reference state is not normalized.'

        if self.symmetry_generators != self.stabilizers:
            # need to update stabilizers in parent class if user decides to fix less stabilizers (e.g. doesn't want
            # to taper all stabilizers). Could be useful in error mitigation strategies
            warnings.warn('the defined symmetry generators have been updated from parent class stabilizers')
            super().__init__(self.symmetry_generators)

        # allow an auxiliary operator (e.g. an Ansatz) to be tapered
        if aux_operator is not None:
            operator_to_taper = aux_operator.copy()
        else:
            operator_to_taper = self.operator.copy()

        # taper the operator via S3Projection.perform_projection
        tapered_operator = self.perform_projection(
            operator=operator_to_taper,
            ref_state=ref_state,
            sector=sector
        )

        # if a reference state was supplied, project it into the stabilizer subspace
        if ref_state is not None:
            self.tapered_ref_state = self.project_state(ref_state)

        return tapered_operator
    
    def project_state(self, state_to_project: QuantumState) -> QuantumState:
        """
        """
        return self._project_state(state_to_project)



"""Main init for package."""
import warnings
warnings.filterwarnings('ignore', module='cotengra')
from symmer.process_handler import process
from symmer.operators  import PauliwordOp, QuantumState
from symmer.projection import QubitTapering, ContextualSubspace, QubitSubspaceManager


from typing import List
from symmer import PauliwordOp, QuantumState
from qiskit import QuantumCircuit
import numpy as np
from qiskit import qasm3
import re

class CircuitSymmerlator:
    """
    Symmer circuit simulator

    Each Clifford gate is expanded as a sequence of pi/2 rotations
    (correct up to some global phase that cancels in the inner product calculation).

    Non-Clifford gates also included, although note the use of these is inefficient.
    
    """
    def __init__(self, n_qubits: int) -> None:
        """
        """
        self.n_qubits = n_qubits
        self.sequence = []
        self.gate_map = {
            'x':self.X,'y':self.Y,'z':self.Z,
            'rx':self.RX,'ry':self.RY,'rz':self.RZ,
            'sx':self.sqrtX,'sy':self.sqrtY,'sz':self.sqrtZ,
            'cx':self.CX,'cy':self.CY,'cz':self.CZ,
            'h':self.H,'s':self.S,'sdg':self.Sdag,
            '':self.R,'t':self.T,'ccx':self.Toffoli,'swap':self.SWAP
        }

    def get_rotation_string(self, pauli: str, indices: List[int]):
        """ Given a Pauli string and list of corresponding qubit indices, return the PauliwordOp
        """
        pauli = list(pauli)
        assert len(pauli)==len(indices), 'Number of Paulis and indices do not match'
        assert set(pauli).issubset({'I','X','Y','Z'}), 'Pauli operators are either I, X, Y or Z.'
        R = ['I']*self.n_qubits
        for i,P in zip(indices, pauli):
            R[i] = P
        return PauliwordOp.from_list([''.join(R)])
    
    def pi_2_multiple(self, multiple: int):
        """ 
        For multiple % 4 = 0,1,2,3 and rotation operator R we have the following behaviour:
            0: +I
            1: +R
            2: -I
            3: -R
        multiplied by the anticommuting component of the target operator.
        """
        return np.pi/2*multiple
    
    #################################################################
    ######################  CLIFFORD GATES  #########################
    #################################################################
    
    def X(self, index: int) -> None:
        """ Pauli iX gate """
        self.sequence.append( (self.get_rotation_string('X', [index]), self.pi_2_multiple(+2)) )
    
    def Y(self, index: int) -> None:
        """ Pauli iY gate """
        self.sequence.append( (self.get_rotation_string('Y', [index]), self.pi_2_multiple(+2)) )
    
    def Z(self, index: int) -> None:
        """ Pauli iZ gate """
        self.sequence.append( (self.get_rotation_string('Z', [index]), self.pi_2_multiple(+2)) )

    def H(self, index: int) -> None:
        """ iH Hadamard gate """
        self.sequence.append( (self.get_rotation_string('Z', [index]), self.pi_2_multiple(+2)) )
        self.sequence.append( (self.get_rotation_string('Y', [index]), self.pi_2_multiple(+1)) )
        
    def S(self, index: int) -> None:
        """ √(+i)S gate """
        self.sequence.append( (self.get_rotation_string('Z', [index]), self.pi_2_multiple(+1)) )

    def Sdag(self, index: int) -> None:
        """ -√(-i)S† gate """
        self.sequence.append( (self.get_rotation_string('Z', [index]), self.pi_2_multiple(+3)) )

    def sqrtX(self, index: int) -> None:
        """ √(iX) gate """
        self.sequence.append( (self.get_rotation_string('X', [index]), self.pi_2_multiple(+1)) )

    def sqrtY(self, index: int) -> None:
        """ √(iY) gate """
        self.sequence.append( (self.get_rotation_string('Y', [index]), self.pi_2_multiple(+1)) )

    def sqrtZ(self, index: int) -> None:
        """ √(iZ) gate """
        self.sequence.append( (self.get_rotation_string('Z', [index]), self.pi_2_multiple(+1)) )

    def CX(self, control: int, target: int) -> None:
        """ √(-i)CX gate """
        self.sequence.append( (self.get_rotation_string('ZX', [control, target]), self.pi_2_multiple(+1)) )
        self.sequence.append( (self.get_rotation_string('ZI', [control, target]), self.pi_2_multiple(+3)) )
        self.sequence.append( (self.get_rotation_string('IX', [control, target]), self.pi_2_multiple(+3)) )
    
    def CY(self, control: int, target: int) -> None:
        """ √(-i)CY gate """
        self.sequence.append( (self.get_rotation_string('ZY', [control, target]), self.pi_2_multiple(+1)) )
        self.sequence.append( (self.get_rotation_string('ZI', [control, target]), self.pi_2_multiple(+3)) )
        self.sequence.append( (self.get_rotation_string('IY', [control, target]), self.pi_2_multiple(+3)) )

    def CZ(self, control: int, target: int) -> None:
        """ √(-i)CZ gate """
        self.sequence.append( (self.get_rotation_string('ZZ', [control, target]), self.pi_2_multiple(+1)) )
        self.sequence.append( (self.get_rotation_string('ZI', [control, target]), self.pi_2_multiple(+3)) )
        self.sequence.append( (self.get_rotation_string('IZ', [control, target]), self.pi_2_multiple(+3)) )

    def SWAP(self, qubit_1: int, qubit_2: int) -> None:
        """ Swap qubits 1 and 2 """
        self.CX(qubit_1,qubit_2)
        self.CX(qubit_2,qubit_1)
        self.CX(qubit_1,qubit_2)
        
    #################################################################
    ####################  NON-CLIFFORD GATES  #######################
    #################### WARNING: INEFFICIENT #######################
    #################################################################
        
    def R(self, pauli: str, indices: List[int], angle: float) -> None:
        """ Arbitrary rotation gate """
        self.sequence.append( (self.get_rotation_string(pauli, indices), -angle) )
    
    def RX(self, index: int, angle: float) -> None:
        """ Pauli X rotation """
        self.R('X', [index], angle)

    def RY(self, index: int, angle: float) -> None:
        """ Pauli Y rotation """
        self.R('Y', [index], angle)

    def RZ(self, index: int, angle: float) -> None:
        """ Pauli Z rotation """
        self.R('Z', [index], angle)

    def T(self, index: int, angle: float) -> None:
        """ T gate """
        raise NotImplementedError()
    
    def Toffoli(self, control_1: int, control_2: int, target: int) -> None:
        """ Doubly-controlled X gate """
        raise NotImplementedError()

    #################################################################
    ######################  GATE EXECUTION  #########################
    #################################################################

    def apply_sequence(self, operator: PauliwordOp) -> PauliwordOp:
        """ Apply the stored sequence of rotations on the input operator
        """
        assert operator.n_qubits == self.n_qubits, 'The operator is defined over a different number of qubits'
        return operator.perform_rotations(self.sequence[::-1])
    
    def evaluate(self, operator: PauliwordOp) -> float:
        """ Evaluate the stored rotations on the input operator
        """
        rotated_op = self.apply_sequence(operator)
        expval = 0
        for rotated_str, coeff in rotated_op.to_dictionary.items():
            if set(rotated_str).issubset({'I','Z'}):
                expval += coeff
        return expval
        
    @classmethod
    def from_qasm(cls, qasm: str, angle_factor: int=1) -> "CircuitSymmerlator":
        """ Initialize the simulator from a QASM circuit
        """
        instructions = qasm.split(';\n')[:-1]
        qasm_version = instructions.pop(0)
        inclusions   = instructions.pop(0)
        registers    = instructions.pop(0)
        # n_qubits     = int(registers.split(' ')[1][2:-1])
        n_qubits =   int(re.findall(r'\d+', registers)[0])

        self = cls(n_qubits)
        pi = np.pi # for evaluating strings like '3*pi/2'
        for step in instructions:
            gate_qubits  = step.split(' ')
            gate = gate_qubits[0]
            qubits = ''.join(gate_qubits[1:])
            qubits = [int(q[2:-1]) for q in qubits.split(',')]
            extract_angle = gate.split('(')
            if len(extract_angle) == 1:
                gate  = extract_angle[0]
                angle = None
            else:
                gate, angle = extract_angle
                angle = eval(angle[:-1])
            if angle is not None:
                self.gate_map[gate](*qubits, angle=angle_factor*angle)
            else:
                self.gate_map[gate](*qubits)
        return self
    
    @classmethod
    def from_qiskit(cls, circuit: QuantumCircuit) -> "CircuitSymmerlator":
        """ Initialize the simulator from a Qiskit QuantumCircuit
        """
        return cls.from_qasm(qasm3.dumps(circuit.reverse_bits()), angle_factor=-1)



import numpy as np
from typing import List, Tuple, Union
from symmer.operators import PauliwordOp, IndependentOp, QuantumState
from symmer.evolution import trotter, Had
from functools import reduce

class S3Projection:
    """ 
    Base class for enabling qubit reduction techniques derived from
    the Stabilizer SubSpace (S3) projection framework, such as tapering
    and Contextual-Subspace VQE. The methods defined herein serve the 
    following purposes:

    - _perform_projection
        Assuming the input operator has been rotated via the Clifford operations 
        found in the above stabilizer_rotations method, this will effect the 
        projection onto the corresponding stabilizer subspace. This involves
        droping any operator terms that do not commute with the rotated generators
        and fixing the eigenvalues of those that do consistently.
    - perform_projection
        This method wraps _perform_projection but provides the facility to insert
        auxiliary rotations (that need not be Clifford). This is used in CS-VQE
        to implement unitary partitioning where necessary. 

    Attributes:
        rotated_flag (bool): If True, the operator is rotated. By default it is set to 'False'.
    """
    rotated_flag = False

    def __init__(self,
                stabilizers: IndependentOp
                ) -> None:
        """
        - eigenvalues: The list of eigenvalue assignments to complement the stabilizers.
        
        - target_sqp: The target single-qubit Pauli (X or Z) that we wish to rotate onto.
        - fix_qubits: Manually overrides the qubit positions selected in stabilizer_rotations, although the rotation procedure can be a bit unpredictable so take care!

        Args:
            stabilizers (IndependentOp): A list of stabilizers that should be enforced, given as Pauli strings.
        """
        self.stabilizers = stabilizers
        
    def _perform_projection(self, 
            operator: PauliwordOp,
            #sym_sector: Union[List[int], np.array]
        ) -> PauliwordOp:
        """ 
        Method for projecting an operator over fixed qubit positions 
        stabilized by single Pauli operators (obtained via Clifford operations).

        Args:
            operator (PauliwordOp): Operator to be projected over fixed qubit positions stabilized by single Pauli operators.

        Returns:
            PauliwordOp representing the projection of input operator.
        """
        assert(operator.n_qubits == self.stabilizers.n_qubits), 'The input operator does not have the same number of qubits as the stabilizers'
        assert(self.rotated_flag), 'The operator has not been rotated - intended for use with perform_projection method'
        self.rotated_flag = False
        
        # remove terms that do not commute with the rotated stabilizers
        commutes_with_all_stabilizers = np.all(operator.commutes_termwise(self.rotated_stabilizers), axis=1)
        op_anticommuting_removed = operator.symp_matrix[commutes_with_all_stabilizers]
        cf_anticommuting_removed = operator.coeff_vec[commutes_with_all_stabilizers]

        # determine sign flipping from eigenvalue assignment
        # currently ill-defined for single-qubit Y stabilizers
        stab_symp_indices  = np.where(self.rotated_stabilizers.symp_matrix)[1]
        eigval_assignment = op_anticommuting_removed[:,stab_symp_indices]*self.rotated_stabilizers.coeff_vec
        eigval_assignment[eigval_assignment==0]=1 # 0 entries are identity, so fix as 1 in product
        coeff_sign_flip = cf_anticommuting_removed*(np.prod(eigval_assignment, axis=1)).T

        # the projected Pauli terms:
        unfixed_XZ_indices = np.hstack([self.free_qubit_indices,
                                        self.free_qubit_indices+operator.n_qubits])
        projected_symplectic = op_anticommuting_removed[:,unfixed_XZ_indices]

        # there may be duplicate rows in op_projected - these are identified and
        # the corresponding coefficients collected in the cleanup method
        if projected_symplectic.shape[1]:
            return PauliwordOp(projected_symplectic, coeff_sign_flip).cleanup()
        else:
            return PauliwordOp(np.array([], dtype=bool), [np.sum(coeff_sign_flip)])
            
    def perform_projection(self,
            operator: PauliwordOp,
            ref_state: Union[List[int], np.array]=None,
            sector: Union[List[int], np.array]=None
        ) -> PauliwordOp:
        """ 
        Input a PauliwordOp and returns the reduced operator corresponding 
        with the specified stabilizers and eigenvalues.
        
        insert_rotation allows one to include supplementary Pauli rotations
        to be performed prior to the stabilizer rotations, for example 
        unitary partitioning in CS-VQE.

        Args: 
            operator (PauliwordOp): Operator projected over fixed qubit positions stabilized by single Pauli operators.
            ref_state (np.array): Reference State. By default, it is set to None.
            sector (np.array): Sector. By default it is set to none. If no sector is provided then a reference state must be given instead.
        Returns:
            Reduced operator corresponding to the given stabilizers and eigenvalues.
        """
        if sector is None and ref_state is not None:
            #assert(ref_state is not None), 'If no sector is provided then a reference state must be given instead'
            self.stabilizers.update_sector(ref_state)
        elif sector is not None:
            self.stabilizers.coeff_vec = np.array(sector, dtype=int)

        self.rotated_stabilizers = self.stabilizers.rotate_onto_single_qubit_paulis()
        self.stab_qubit_indices  = np.where(self.rotated_stabilizers.symp_matrix)[1] % operator.n_qubits
        self.free_qubit_indices  = np.setdiff1d(np.arange(operator.n_qubits),self.stab_qubit_indices)

        # perform the full list of rotations on the input operator...
        if len(self.stabilizers.stabilizer_rotations) > 0:
            op_rotated = operator.perform_rotations(self.stabilizers.stabilizer_rotations)
        else:
            op_rotated = operator
        
        self.rotated_flag = True
        # ...and finally perform the stabilizer subspace projection
        return self._perform_projection(operator=op_rotated)
    
    def _project_state(self, state: QuantumState) -> QuantumState:
        """ 
        Project a state into the stabilizer subspace.

        Args:
            state (QuantumState): The state which has to be projected into the stabilizer subspace.

        Returns: 
            Projection of input QuantumState into the stabilizer subspace.
        """
        transformation_list = []
        # Hadamards where rotated onto Pauli X operators
        transformation_list += [
            Had(self.stabilizers.n_qubits, i) for i in np.where(
                np.sum(
                    self.stabilizers.rotate_onto_single_qubit_paulis().X_block & 
                    ~self.stabilizers.rotate_onto_single_qubit_paulis().Z_block,
                    axis=0
            )
                )[0]
        ]
        # Projections onto the stabilizer subspace
        transformation_list += list(map(lambda x:(x**2 + x)*.5,self.stabilizers.rotate_onto_single_qubit_paulis()))
        # Rotations mapping stabilizers onto single-qubit Pauli operators
        transformation_list += list(map(lambda s:trotter(s[0]*(np.pi/4*1j)), self.stabilizers.stabilizer_rotations))
        # Product over the transformation list yields final transformation operator
        transformation = reduce(lambda x,y:x*y, transformation_list)
        # apply transformation to the reference state
        transformed_state = transformation * state
        # drop stabilized qubit positions and sum over potential duplicates
        return QuantumState(
            transformed_state.state_matrix[:, self.free_qubit_indices], 
            transformed_state.state_op.coeff_vec
        ).cleanup(zero_threshold=1e-12)


import numpy as np
from symmer.operators import PauliwordOp, IndependentOp, NoncontextualOp, QuantumState
from symmer.projection.utils import (
    update_eigenvalues, StabilizerIdentification, ObservableBiasing, stabilizer_walk,
    # get_noncon_generators_from_commuting_stabilizers
)
from symmer.projection import S3Projection
from symmer.evolution import trotter
from typing import List, Union, Optional

class ContextualSubspace(S3Projection):
    """ 
    Class for performing contextual subspace methods as per https://quantum-journal.org/papers/q-2021-05-14-456/.
    Reduces the number of qubits in the problem while aiming to control the systematic error incurred along the way.

    This class handles the following:
    1. Identify a set of operators one wishes to enforce as stabilizers over the contextual subspace,
        one might think ofthese as 'pseudo-symmetries', as opposed to the true, physical symmetries of qubit tapering.
    2. Construct a noncontextual Hamiltoinian that respects the stabilizers selecting in (1),
        the NoncontextualOp class handles the decomposition into a generating set and classical optimization over the noncontextual objective function 
    NOTE: the order in which (1) and (2) are performed depends on the noncontextual strategy specified
    3. Apply unitary partitioning (either sequence of rotations or linear combination of unitaries) to collapse noncontextual cliques
    
    The remaining steps are handled by the parent S3Projection class:
    4. rotate each stabilizer onto a single-qubit Pauli operator, 
    5. drop the corresponding qubits from the Hamiltonian whilst
    6. fixing the +/-1 eigenvalues
    """
    name = 'contextual_subspace' # for reference in QubitSubspaceManager

    def __init__(self,
            operator: PauliwordOp,
            noncontextual_strategy: str = 'diag',
            noncontextual_solver: str = 'brute_force',
            unitary_partitioning_method: str = 'seq_rot',
            reference_state: Union[np.array, QuantumState] = None,
            noncontextual_operator: NoncontextualOp = None
        ):
        """
        When passing in a noncontextual_operator if noncontextual_strategy set to be 'solved' then noncontextual_solver will NOT be run.

        Args: 
            operator(PauliwordOp): Operator one wishes to enforce as stabilizers over the contextual subspace.
            noncontextual_strategy (str): Non-Contextual Strategy to be applied. Its default value is'diag'.
            noncontextual_solver (str): Non-contextual solver to be applied. Its default value is'brute_force'.
            unitary_partitioning_method (str): Unitary Partitioning Method to be applied. Its default value is'seq_rot'.
            reference_state (QuantumState): Reference State. By default, it is set to None.
            noncontextual_operator (NoncontextualOp): Non-contextual Operator. By default, it is set to None.    
        """
        # noncontextual startegy will have the form x_y, where x is the actual strategy
        # and y is some supplementary method indicating a sorting key such as magnitude
        if reference_state is None or isinstance(reference_state, QuantumState):
            self.ref_state = reference_state
        else:
            self.ref_state = QuantumState(reference_state)            
        extract_noncon_strat = noncontextual_strategy.split('_')
        self.nc_strategy = extract_noncon_strat[0]
        self.noncontextual_solver = noncontextual_solver
        self.unitary_partitioning_method = unitary_partitioning_method

        # With the exception of the StabilizeFirst noncontextual strategy, here we build
        # the noncontextual Hamiltonian in line with the specified strategy
        self.operator = operator
        if noncontextual_operator is None and self.nc_strategy != 'StabilizeFirst':
            self.noncontextual_operator = NoncontextualOp.from_hamiltonian(
                operator, strategy=noncontextual_strategy
            )
        else:
            self.noncontextual_operator = noncontextual_operator
        self.unitary_partitioning_method = unitary_partitioning_method
        self._noncontextual_update()

    def manual_stabilizers(self, S: Union[List[str], IndependentOp]) -> None:
        """ 
        Specify a set of operators to enforce manually.

        Args:
            S (IndependentOp): Set of Stablizers
        """
        if isinstance(S, list):
            S = IndependentOp.from_list(S)
        self.n_qubits_in_subspace = self.operator.n_qubits - S.n_terms
        if self.n_qubits_in_subspace == 0:
            self.return_NC = True
        else:
            self.return_NC = False
        self.stabilizers = S
        self._prepare_stabilizers()

    def update_stabilizers(self, 
            n_qubits: int, 
            strategy: str = 'aux_preserving',
            aux_operator: PauliwordOp = None,
            HF_array: np.array = None,
            use_X_only: bool = True
        ) -> None:
        """ 
        Update the stabilizers that will be used for the subspace projection.

        Args:
            n_qubits (int): Number of Qubits
            strategy (str): Strategy to be applied. It's default value is 'aux_preserving'.
            aux_operator (PauliwordOp): Auxiliary operator. By default, it is set to None.
            HF_array (np.array): Hartree-Fock state. By default, it is set to None.
            use_X_only (bool): Default value is 'True'.
        """
        assert(n_qubits<=self.operator.n_qubits), (
            'Cannot define a contextual subspace larger than the base Hamiltonian'
        )
        if n_qubits == 0:
            n_qubits = 1
            self.return_NC = True
        else:
            self.return_NC = False

        if n_qubits == self.operator.n_qubits:
            self.stabilizers = None
        else:
            if strategy == 'aux_preserving':
                S = self._aux_operator_preserving_stabilizer_search(
                    n_qubits=n_qubits, aux_operator=aux_operator, use_X_only=use_X_only
                )
            elif strategy == 'random':
                S = self._random_stabilizers(
                    n_qubits=n_qubits
                )
            elif strategy == 'HOMO_LUMO_biasing':
                S = self._HOMO_LUMO_biasing(
                    n_qubits=n_qubits, HF_array=HF_array, 
                    weighting_operator=aux_operator, use_X_only=use_X_only
                )
            else:
                raise ValueError('Unrecognised stabilizer search strategy.')

            self.n_qubits_in_subspace = self.operator.n_qubits - S.n_terms
            self.stabilizers = S
            self._prepare_stabilizers()

    def _noncontextual_update(self):
        """ 
        To be executed each time the noncontextual operator is updated.
        """
        if self.noncontextual_operator is not None:
            self.noncontextual_operator.up_method = self.unitary_partitioning_method
            self.contextual_operator = self.operator - self.noncontextual_operator
            if self.contextual_operator.n_terms == 0:
                raise ValueError('The Hamiltonian is noncontextual, the contextual subspace is empty.')
            if self.nc_strategy != "solved":
                self.noncontextual_operator.solve(
                    strategy=self.noncontextual_solver, 
                    ref_state=self.ref_state
                )
            else:
                self.noncontextual_operator.update_clique_representative_operator()
            self.n_cliques = self.noncontextual_operator.n_cliques
        
    def _aux_operator_preserving_stabilizer_search(self,
            n_qubits: int,
            aux_operator: PauliwordOp,
            use_X_only: bool = True
        ) -> IndependentOp:
        """
        Choose stabilizers that preserve some auxiliary operator.
        This could be an Ansatz operator such as UCCSD, for example.

        Args:
            n_qubits (int): Number of Qubits
            aux_operator (PauliwordOp): Auxiliary operator.
            use_X_only (bool): Default value is 'True'.

        Returns:
            S (IndependentOp):  Stablizer that preserves the passed auxiliary operator.
        """
        if aux_operator is None:
            if self.nc_strategy == 'StabilizeFirst':
                aux_operator = self.operator
            else:
                aux_operator = self.contextual_operator

        SI = StabilizerIdentification(aux_operator, use_X_only=use_X_only)
        S = SI.symmetry_generators_by_subspace_dimension(n_qubits)

        return S

    def _HOMO_LUMO_biasing(self,
            n_qubits: int,
            HF_array: np.array,
            weighting_operator: PauliwordOp = None,
            use_X_only:bool = True
        ) -> IndependentOp:
        """ 
        Bias the Hamiltonian with respect to the HOMO-LUMO gap 
        and preserve terms in the resulting operator as above.

        Args:
            n_qubits (int): Number of Qubits
            HF_array (np.array): Hartree-Fock state
            weighting_operator (PauliwordOp): Weighting Operator. By default, it is set to None.
            use_X_only (bool): Default value is 'True'.

        Returns:
            S (IndependentOp): Set of Stablizers
        """
        assert(HF_array is not None), 'Must supply the Hartree-Fock state for this strategy'
        
        OB = ObservableBiasing(
            base_operator=self.operator, 
            HOMO_LUMO_gap=np.where(np.asarray(HF_array==0).reshape(-1))[0][0]-.5 # currently assumes JW mapping!
        )
        S = stabilizer_walk(
            n_sim_qubits=n_qubits, 
            biasing_operator=OB, 
            weighting_operator=weighting_operator,
            use_X_only=use_X_only
        )
        return S

    def _random_stabilizers(self, 
            n_qubits: int
        )  -> IndependentOp:
        """ 
        Generate a random set of stabilizers.

        Args:
            n_qubits (int): Number of Qubits
        
        Returns: 
            S (IndependentOp): Random set of Stablizers
        """
        # TODO better approach that does not rely on this *potentially infinite* while loop!
        found_stabilizers=False
        while not found_stabilizers:
            try:
                S = PauliwordOp.random(
                    self.operator.n_qubits,
                    self.operator.n_qubits-n_qubits,
                    diagonal=True
                )
                S.coeff_vec[:] = 1
                S = IndependentOp.from_PauliwordOp(S)
                found_stabilizers = True
            except:
                pass
        
        return S

    def _prepare_stabilizers(self) -> None:
        """ 
        Prepare the chosen stabilizers for projection into the contextual subspace.
        This includes eigenvalue assignment (obtained from the solution of the noncontextual Hamiltonian), and application of unitary partitioning if enforcing a clique element.
        """
        self.S3_initialized = False
        #the StabilizeFirst strategy differs from the others in that the noncontextual
        #Hamiltonian is constructed AFTER selecting stabilizers, which is what we do here:
        if self.nc_strategy == 'StabilizeFirst':
            self.noncontextual_operator = NoncontextualOp._from_stabilizers_noncontextual_op(
                H=self.operator, stabilizers=self.stabilizers, use_jordan_product=False
            )
            self._noncontextual_update()

        if self.noncontextual_operator.n_cliques > 0:
            # mask stabilizers that lie within one of the noncontextual cliques
            clique_commutation = self.stabilizers.commutes_termwise(self.noncontextual_operator.clique_operator)
            mask_which_clique = np.all(clique_commutation, axis=0)
        else:
            mask_which_clique = []

        if ~np.all(mask_which_clique):
            # we may only enforce stabilizers that live within the same clique, not accross them:
            assert(sum(mask_which_clique)==1), (
                'Cannot enforce stabilizers from different cliques since '+
                'unitary partitioning collapses onto just one of them.'
            )
            # generate the unitary partitioning rotations that map onto the 
            # clique representative correpsonding with the given stabilizers
            self.noncontextual_operator.update_clique_representative_operator(
                clique_index=int(np.where(mask_which_clique)[0][0])
            )
            # add the clique representative to the noncontextual generators in order to 
            # update the eigenvalue assignments of the chosen stablizers so they are 
            # consistent with the noncontextual ground state configuration - this is 
            # G U {RARdag} in the original CS-VQE notation. 
            augmented_generators = (
                IndependentOp(self.noncontextual_operator.mapped_clique_rep.symp_matrix, [-1]) + 
                self.noncontextual_operator.symmetry_generators
            )
            # given these new generators, we reconstruct the given stabilizers to identify
            # the correct subspace corresponding with the noncontextual ground state (nu, r)
            update_eigenvalues(generators=augmented_generators, stabilizers=self.stabilizers)
            self.perform_unitary_partitioning = True
        else:
            update_eigenvalues(
                generators=self.noncontextual_operator.symmetry_generators, 
                stabilizers=self.stabilizers
            )
            self.perform_unitary_partitioning = False

    def project_onto_subspace(self, operator_to_project:PauliwordOp=None) -> PauliwordOp:
        """ 
        Projects with respect to the current stabilizers; these are 
        updated using the ContextualSubspace.update_stabilizers method.

        Args:
            operator_to_project (PauliwordOp): Operator to be projected. By default, it is set to None.

        Returns:
            Projection of operator passed.
        """
        # if not supplied with an alternative operator for projection, use the internal operator 
        if operator_to_project is None:
            operator_to_project = self.operator.copy() 
        # if there are no stabilizers, return the input operator
        if self.stabilizers is None:
            return operator_to_project   
        # instantiate the parent S3Projection class that handles the subspace projection
        super().__init__(self.stabilizers)
        self.S3_initialized = True
        # perform unitary partitioning
        if self.perform_unitary_partitioning:
            rotated_op = operator_to_project.perform_rotations(self.noncontextual_operator.unitary_partitioning_rotations)
        else:
            rotated_op = operator_to_project
        # finally, project the operator before returning
        cs_operator = self.perform_projection(rotated_op)

        if self.return_NC:
            assert cs_operator.n_qubits == 1, 'Projected operator consists of more than one qubit.'
            cs_operator = NoncontextualOp.from_PauliwordOp(cs_operator)
            cs_operator.solve()
            return cs_operator.energy
        else:
            return cs_operator

    def project_state(self, 
            state_to_project: QuantumState = None
        ) -> QuantumState:
        """ 
        Project a QuantumState into the contextual subspace

        Args:
            state_to_project (QuantumState): Quantum State to be projected into the contextual subspace. By default, it is set to None.

        Reutrns:
            Projection of passed Quantum State into the contextual subspace.
        """
        # if there are no stabilizers, return the input QuantumState
        if self.stabilizers is None:
            return state_to_project
        
        assert self.S3_initialized, 'Must first project an operator into the contextual subspace via the project_onto_subspace method'
        # can provide an auxiliary state to project, although not in general scalable
        if state_to_project is None:
            assert self.ref_state is not None, 'Must provide a state to project into the contextual subspace'
            state_to_project = self.ref_state

        if self.perform_unitary_partitioning:
            if self.noncontextual_operator.unitary_partitioning_rotations == []:
                rotation = PauliwordOp.from_list(['I'*self.operator.n_qubits])
            else:
                rotation_generator = sum([R*angle*.5*1j for R,angle in self.noncontextual_operator.unitary_partitioning_rotations])
                rotation = trotter(rotation_generator)
            return self._project_state(rotation * state_to_project)
        else:
            return self._project_state(state_to_project)


import numpy as np
from symmer.operators import PauliwordOp
from symmer.evolution import trotter

#############################################
# Gate library decomposed into PauliwordOps #
#############################################

def I(n_qubits:int) -> PauliwordOp:
    """
    Identity gate

    Args:
        n_qubits (int): Number of qubits

    Returns:
        PauliwordOp representing the identity operation ('I') applied to a system of 'n_qubits'.
    """
    return PauliwordOp.from_dictionary({'I'*n_qubits:1})

def X(n_qubits:int, index:int) -> PauliwordOp:
    """
    Pauli X gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'X' has to be applied.

    Returns:
        PauliwordOp representing the Pauli 'X' operator applied to the specified qubit index while leaving the other qubits unchanged.
    """
    X_str = ['I']*n_qubits; X_str[index] = 'X'
    return PauliwordOp.from_dictionary({''.join(X_str):1})

def Y(n_qubits:int, index:int) -> PauliwordOp:
    """
    Pauli Y gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'Y' has to be applied.

    Returns:
        PauliwordOp representing the Pauli 'Y' operator applied to the specified qubit index while leaving the other qubits unchanged.
    """
    Y_str = ['I']*n_qubits; Y_str[index] = 'Y'
    return PauliwordOp.from_dictionary({''.join(Y_str):1})

def Z(n_qubits:int, index:int) -> PauliwordOp:
    """
    Pauli Z gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'Z' has to be applied.

    Returns:
        PauliwordOp representing the Pauli 'Z' operator applied to the specified qubit index while leaving the other qubits unchanged.
    """
    Z_str = ['I']*n_qubits; Z_str[index] = 'Z'
    return PauliwordOp.from_dictionary({''.join(Z_str):1})

def Had(n_qubits:int, index:int) -> PauliwordOp:
    """ 
    Hadamard gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'H' has to be applied.

    Returns:
        PauliwordOp representing the 'H' operator applied to the specified qubit index while leaving the other qubits unchanged.
    """
    return (
        Z(n_qubits, index).multiply_by_constant(1/np.sqrt(2))+
        X(n_qubits, index).multiply_by_constant(1/np.sqrt(2))
    )

def CZ(n_qubits:int, control:int, target:int) -> PauliwordOp:
    """ 
    Controlled Z gate

    Args:
        n_qubits (int): Number of qubits.
        control (int): Qubit index at which will act as a control qubit.
        target (int): Qubit index at which the operation 'Z' has to be applied if the control qubit is in |1> state.

    Returns:
        PauliwordOp representing the Controlled-Z (CZ) gate applied to the specified control and target qubits in a system of 'n_qubits' qubits.
    """
    ZI = Z(n_qubits, control)
    IZ = Z(n_qubits, target)
    ZZ = ZI * IZ
    
    CZ_exp = (ZZ - IZ - ZI).multiply_by_constant(np.pi/4)
    CZ = trotter(CZ_exp.multiply_by_constant(1j), trotnum=1).multiply_by_constant(np.sqrt(1j))
    return CZ

def CX(n_qubits:int, control:int, target:int) -> PauliwordOp:
    """ 
    Controlled X gate

    Args:
        n_qubits (int): Number of qubits.
        control (int): Qubit index at which will act as a control qubit.
        target (int): Qubit index at which the operation 'X' has to be applied if the control qubit is in |1> state.

    Returns:
        PauliwordOp representing the Controlled-X (CX) gate applied to the specified control and target qubits in a system of 'n_qubits' qubits.
    """
    _Had = Had(n_qubits, target)
    return _Had * CZ(n_qubits, control, target) * _Had

def CY(n_qubits:int, control:int, target:int) -> PauliwordOp:
    """ 
    Controlled Y gate

    Args:
        n_qubits (int): Number of qubits.
        control (int): Qubit index at which will act as a control qubit.
        target (int): Qubit index at which the operation 'X' has to be applied if the control qubit is in |1> state.

    Returns:
        PauliwordOp representing the Controlled-X (CX) gate applied to the specified control and target qubits in a system of 'n_qubits' qubits.
    """
    _Had = Had(n_qubits, target)
    _S   = S(n_qubits, target)
    return _S * _Had * CZ(n_qubits, control, target) * _Had * _S.dagger

def RX(n_qubits:int, index:int, angle:float) -> PauliwordOp:
    """ 
    Rotation-X gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'RX' has to be applied.
        angle (float): The angle by which the qubit state has to be rotated around the X-axis.

    Returns:
        PauliwordOp representing the rotation around the X-axis (RX) gate by the specified angle applied to a specific qubit in a system of 'n_qubits'.
    """
    return trotter(X(n_qubits, index).multiply_by_constant(1j*angle/2))

def RY(n_qubits:int, index:int, angle:float) -> PauliwordOp:
    """ 
    Rotation-Y gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'RY' has to be applied.
        angle (float): The angle by which the qubit state has to be rotated around the Y-axis.

    Returns:
        PauliwordOp representing the rotation around the Y-axis (RY) gate by the specified angle applied to a specific qubit in a system of 'n_qubits'.
    """
    return trotter(Y(n_qubits, index).multiply_by_constant(1j*angle/2))

def RZ(n_qubits:int, index:int, angle:float) -> PauliwordOp:
    """ 
    Rotation-Z gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'RZ' has to be applied.
        angle (float): The angle by which the qubit state has to be rotated around the Z-axis.

    Returns:
        PauliwordOp representing the rotation around the Z-axis (RZ) gate by the specified angle applied to a specific qubit in a system of 'n_qubits'.
    """
    return trotter(Z(n_qubits, index).multiply_by_constant(1j*angle/2))

def U1(n_qubits:int, index:int, angle:float) -> PauliwordOp:
    """ 
    U1 gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the phase shift is to be introduced to the qubit state.
        angle (float): Phase angle.

    Returns:
        PauliwordOp representing the U1 gate with phase 'angle' applied to a specific qubit in a system of 'n_qubits'.
    """
    return RZ(n_qubits, index, angle).multiply_by_constant(np.exp(1j*angle/2))

def S(n_qubits:int, index:int) -> PauliwordOp:
    """
    S gate

    Args:
        n_qubits (int): Number of qubits.
        index (int): Qubit index at which the operation 'S' has to be applied.

    Returns:
        PauliwordOp representing the 'S' operator applied to the specified qubit index while leaving the other qubits unchanged.
    """
    return RZ(n_qubits,index,-np.pi/2).multiply_by_constant(np.sqrt(1j))
